import 'dart:math' as math;
import 'package:dolphinsr_dart/src/models.dart';
import 'package:dolphinsr_dart/src/utils.dart';
import 'package:test/test.dart';

import 'dates.dart';

final master = generateId();

String generateId() {
  return math.Random().nextInt(666).toString();
}

Review makeReview(DateTime ts) {
  return Review(
    master: generateId(),
    combination: Combination(front: [0], back: [1]),
    ts: ts,
    rating: Rating.Easy,
  );
}

final List<Review> reviews = [
  today,
  todayAt3AM,
  laterToday,
  laterTmrw,
  laterInTwoDays,
  laterInFourDays,
].map(makeReview).toList();

void main() {
  test('should add a rounded interval to the lastReviewed, set at 3am', () {
    var id = generateId();
    var combination = Combination(front: [0], back: [1]);
    var stateCard = ReviewingCardState(
        master: id,
        combination: combination,
        lapses: 0,
        lastReviewed: today,
        factor: 1000,
        interval: 13.3);

    var due = calculateDueDate(stateCard);

    var expectedDate = todayAt3AM.add(Duration(days: 14));
    expect(due, equals(expectedDate));
  });
  test('should return later for cards that are reviewing and not yet due', () {
    var id = generateId();
    var combination = Combination(front: [0], back: [1]);

    var stateCard = ReviewingCardState(
        master: id,
        combination: combination,
        lapses: 0,
        lastReviewed: today,
        factor: 1000,
        interval: 13.3);

    var computeSchedule = computeScheduleFromCardState(stateCard, laterTmrw);

    expect(computeSchedule, equals('later'));
  });

  test('should return due for cards that are reviewing and due within the day',
      () {
    var id = generateId();
    var combination = Combination(front: [0], back: [1]);

    var stateCard = ReviewingCardState(
        master: id,
        combination: combination,
        lapses: 0,
        lastReviewed: today,
        factor: 1000,
        interval: 13.3);

    var computeSchedule = computeScheduleFromCardState(
        stateCard, todayAt3AM.add(Duration(days: 14)));

    expect(computeSchedule, equals('due'));

    computeSchedule = computeScheduleFromCardState(
        stateCard, laterToday.add(Duration(days: 14)));
    expect(computeSchedule, equals('due'));
  });

  test('should return overdue for cards that reviewing and due before the day',
      () {
    var id = generateId();
    var combination = Combination(front: [0], back: [1]);

    var stateCard = ReviewingCardState(
        master: id,
        combination: combination,
        lapses: 0,
        lastReviewed: today,
        factor: 1000,
        interval: 13.3);

    var computeSchedule = computeScheduleFromCardState(
        stateCard, todayAt3AM.add(Duration(days: 15)));

    expect(computeSchedule, equals('overdue'));

    computeSchedule = computeScheduleFromCardState(
        stateCard, laterToday.add(Duration(days: 15)));
    expect(computeSchedule, equals('overdue'));
  });

  test('should return an empty schedule when passed an empty state', () {
    var emptyState = makeEmptyState();

    expect(computeCardsSchedule(emptyState, today).learning!.length, equals(0));
    expect(computeCardsSchedule(emptyState, today).later!.length, equals(0));
    expect(computeCardsSchedule(emptyState, today).due!.length, equals(0));
    expect(computeCardsSchedule(emptyState, today).overdue!.length, equals(0));
  });

  test('should a sorted list of cards when passed cards in multiple states',
      () {
    var emptyState = makeEmptyState();

    var id = generateId();
    var combination = Combination(front: [0], back: [1]);

    var dueLater = ReviewingCardState(
        master: id,
        combination: combination,
        lapses: 0,
        lastReviewed: laterTmrw,
        factor: 1000,
        interval: 13.3);
    var dueNow = ReviewingCardState(
        master: id,
        combination: Combination(front: [0, 1], back: [1]),
        lapses: 0,
        lastReviewed: laterTmrw,
        factor: 1000,
        interval: 0);

    var overDue = ReviewingCardState(
        master: id,
        combination: Combination(front: [1], back: [0]),
        lapses: 0,
        lastReviewed: today,
        factor: 1000,
        interval: 0);

    var learning = LearningCardState(
      master: id,
      combination: Combination(front: [1, 0], back: [0]),
      lastReviewed: today,
      consecutiveCorrect: 0,
    );

    var lapsed = LapsedCardState(
        master: id,
        combination: Combination(front: [1, 0], back: [0, 1]),
        lastReviewed: today,
        consecutiveCorrect: 0,
        factor: 1000,
        interval: 0,
        lapses: 1);

    [dueLater, dueNow, overDue, learning, lapsed].forEach((cardState) {
      emptyState.cardStates[getCardIdFromCardState(cardState)] = cardState;
    });

    var s = computeCardsSchedule(emptyState, laterTmrw);

    expect(s.learning!.length, equals(2));
    expect(s.learning![1].uniqueId, equals(getCardIdFromCardState(lapsed)));
    expect(s.learning![0].uniqueId, equals(getCardIdFromCardState(learning)));
    expect(s.later![0].uniqueId, equals(getCardIdFromCardState(dueLater)));
    expect(s.due![0].uniqueId, equals(getCardIdFromCardState(dueNow)));
    expect(s.overdue![0].uniqueId, equals(getCardIdFromCardState(overDue)));
  });

  test('PickMostDue should return null when passed an empty schedule and state',
      () {
    var emptyState = makeEmptyState();
    var s = computeCardsSchedule(emptyState, today);
    final pick = pickMostDue(s, emptyState);
    expect(pick, isNull);
  });

  test(
      'PickMostDue should return the learning card reviewed most recently if two learning cards are in the deck',
      () {
    var emptyState = makeEmptyState();

    var id = generateId();
    var combination = Combination(front: [0], back: [1]);

    var dueLater = ReviewingCardState(
        master: id,
        combination: combination,
        lapses: 0,
        lastReviewed: laterTmrw,
        factor: 1000,
        interval: 13.3);
    var dueNow = ReviewingCardState(
        master: id,
        combination: Combination(front: [0, 1], back: [1]),
        lapses: 0,
        lastReviewed: laterTmrw,
        factor: 1000,
        interval: 0);

    var overDue = ReviewingCardState(
        master: id,
        combination: Combination(front: [1], back: [0]),
        lapses: 0,
        lastReviewed: today,
        factor: 1000,
        interval: 0);

    var learning = LearningCardState(
      master: id,
      combination: Combination(front: [1, 0], back: [0]),
      lastReviewed: today,
      consecutiveCorrect: 0,
    );

    var lapsed = LapsedCardState(
        master: id,
        combination: Combination(front: [1, 0], back: [0, 1]),
        lastReviewed: laterTmrw,
        consecutiveCorrect: 0,
        factor: 1000,
        interval: 0,
        lapses: 1);

    [dueLater, dueNow, overDue, learning, lapsed].forEach((cardState) {
      emptyState.cardStates[getCardIdFromCardState(cardState)] = cardState;
    });

    var s = computeCardsSchedule(emptyState, laterTmrw);
    var pick = pickMostDue(s, emptyState)!;
    expect(pick.uniqueId, equals(getCardIdFromCardState(learning)));
  });
}
